#!/usr/bin/env ruby
$KCODE = 'e'

=begin
Trinity Project - extensions for werewolves.

Copyright (C)2006 miyamo

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to
deal in the Software without restriction, including without limitation the
rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
sell copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.

$Id: index.cgi,v 1.24.2.7 2006/12/12 04:22:47 miyamo Exp $

=end

require 'nkf'
require 'time'
require 'cgi'
require 'erb'
require 'net/http'
require 'open-uri'
require 'logger'
require 'uri'
require 'strscan'

module Software
	Name = 'Trinity'
	Code = 'Trinity'
	Page = 'http://pixygarden.net/ys/wolf/?p=Trinity'
	Version = '20061212'
	TemplateDir = 'view/'
	PrivateDir = 'private/'
	ImageDir = PrivateDir + 'img/'
	LogDir = PrivateDir + 'log/'
	LogFile = LogDir + Code + '.log'
	ConfigFile = PrivateDir + 'user.cfg'
end

Dir.mkdir(Software::LogDir) unless File.exist?(Software::LogDir)
$logger = Logger.new(Software::LogFile, 1)
$logger.progname = Software::Code
$logger.level = Logger::DEBUG

class File
	def File.touch(fp)
		File.open(fp, 'w') {|fh| fh.write '' } unless File.exist?(fp)
	end

	def File.dump(fp, content)
		File.open(fp, 'w') {|fh| fh.write content }
	end
end

class CGI
	def get(key)
		self.params[key][0] || nil
	end

	def get_values(ary)
		ret = []
		ary.each {|key| ret << self.params[key][0] || nil }
		ret
	end

	def get_values_with_default(ary, hash)
		ret = []
		ary.each {|key|
			if self.params.has_key?(key)
				ret << self.params[key][0]
			else
				ret << hash[key]
			end
		}
		$logger.debug(ret.join(' '))
		ret
	end

	def CGI.url(msg)
		msg.gsub(/((http|https):\/\/[^ۡ<\]\s\n]+)/, '<a href="\1">\1</a>')
	end

	def CGI.sprint(str)
		c = self.new
		c.out('text/html') { str }
		exit
	end
end

module Trinity

	############################################################ IMPL

	include Software

	# ERB ƥץ졼ȥեɤ߹
	module Template
		def self.[](name)
			File.read(Software::TemplateDir + name + '.rhtml')
		end
	end

	# ƥ饤֥ɤ߹(Const)
	module Library
		def self.load(path)
			module_eval(File.read(path))
		end
	end

	# ե Sync
	class InitImages
		def initialize(version)
			@version = version
		end

		def self.sync(path)
			module_eval(File.read(path))
			new(Software::Version).sync
		end

		def sync
			syncfile = Software::ImageDir + @version + '.done'
			unless File.exist?(syncfile)
				File.touch(syncfile)
				ret = false
				List.each {|uri|
					fn = File.basename(uri)
					unless File.exist?(Software::ImageDir + fn)
						$logger.debug("InitImages#sync: #{fn} for #{@version}")
						http_get(uri, Software::ImageDir + fn)
						ret = true
					end
				}
				ret
			else
				false
			end
		end

		def http_get(uri, fp)
			f = OpenURI.open_uri(
				uri,
				'User-Agent' => "#{Software::Code} (#{Software::Version})"
				)
			File.open(fp, 'wb') {|fh| fh.write f.read }
		end
	end

	#  URI
	class EuroCountryMap < Hash
		attr_reader :name

		def initialize(vid)
			super()
			self['vid'] = vid
			self['vnum'] = vid.sub(/^O/, '')
			self['active'] = true
			self['country'] = 'O'

			base = 'http://euros.sakura.ne.jp/wolf/index.rb?'

			self['uri4get'] = base + "vil_id=#{self['vnum']}"
			self['uri'] = base + "vid=#{self['vnum']}#form"
			self['image'] = './img/'

			ary = base.split('/')
			self['path4head'] = "/#{ary[3]}/#{ary[4]}vil_id=#{self['vnum']}"
			self['host'] = ary[2]
			self['path4post'] = "/#{ary[3]}/#{ary[4].delete('?')}"
			@name = 'euro'
			constmapping
		end

		# parser ǻȤʸ/ɽ
		def constmapping
			self['logkey'] = 'id'
			self['party'] = false
			self['regex_today'] = %r{ <span class="time">(.*?)</span> }
			self['regex_today_days'] = /(\d+)/
			self['regex_today_epi'] = false
			self['regex_today_end'] = false
			self['regex_today_cut'] = %r{<table class="member">.*}m
			self['regex_pageuris_head_end_0'] = \
					%r(.*<a href="\?vil_id=#{self['vnum']}&date=0&log=all"></a> )m
			self['regex_pageuris_head_end_1'] = //
			self['regex_pageuris_foot_start'] = %r( <a href="./index.rb?vil_id=#{self['vnum']}#form">ǿ</a>.*)m
			self['regex_pageuris_uris'] = %r{<a href="\?vil_id=#{self['vnum']}&(date=.*?&log=all)">}
			self['regex_pageuris_ready'] = false
			self['regex_pageuris_party'] = false
			self['regex_pageuris_days'] = /date=(.*?)&log=all/
			self['pageuris_days_succ'] = false
			self['mes_all'] = 'log=all'
			self['tmppage_voicecnt'] = 48
		end
	end

	# Ninjin  URI / ƥ֤ɤ Hash
	class NinjinCountryMap < Hash
		attr_reader :name

		def initialize(vid)
			super()
			self['vid'] = vid
			self['vnum'] = vid.sub(/^[ABCDEF]/, '')
			self['active'] = false

			case vid
			when /^(\w)/ then self['country'] = $1.dup
			when /\d{3}/ then self['country'] = '0'  # ܹ
			else
				self['country'] = '1'  # 
			end

			base = Library::Ninjin::SrvUris[self['country']]
			self['uri4get'] = base + "vid=#{self['vnum']}"
			self['uri'] = base + "vid=#{self['vnum']}#bottom"

			Library::Ninjin::SrvImageUris.each {|k, v|
				self['image'] = v if k.index(self['country'])
			}

			ary = base.split('/')
			self['path4head'] = "/#{ary[3]}/#{ary[4]}vid=#{self['vnum']}&cmd=rss"
			if self['country'] =~ /[CF]/
				self['host'] = ary[2]
				self['path4post'] = "/#{ary[3]}/#{ary[4].delete('?')}"
				self['active'] = true
			end
			@name = 'ninjin'
			constmapping
		end

		# parser ǻȤʸ/ɽ
		def constmapping
			self['logkey'] = 'time'
			self['party'] = 'party'
			self['regex_today'] = %r{<span class="time">(.*?)</span></p>}
			self['regex_today_days'] = /(\d)/
			self['regex_today_epi'] = /ԥ/
			self['regex_today_end'] = /λ/
			self['regex_today_cut'] = //
			self['regex_pageuris_head_end_0'] = \
					%r(.*<!---<a href="index\.rb\?vid=#{self['vnum']}&amp;cmd=mb">simple</a>--->\n\n\n)m
			self['regex_pageuris_head_end_1'] = \
					%r(.*<a href="index\.rb\?vid=#{self['vnum']}&amp;cmd=rss">rss</a>\n)m
			self['regex_pageuris_foot_start'] = %r(\n+<div class="message">\n.*)m
			self['regex_pageuris_uris'] = %r{<a href="index\.rb\?vid=#{self['vnum']}&amp;(.*?)">}
			self['regex_pageuris_ready'] = /ready/
			self['regex_pageuris_party'] = /party/
			self['regex_pageuris_days'] = /(\d)$/
			self['pageuris_days_succ'] = true
			self['mes_all'] = 'mes=all'
			self['tmppage_voicecnt'] = 8
		end
	end

	class Downloader
		attr_reader :result
	
		def initialize(vid)
			case vid
			when /^O/
				@map = EuroCountryMap.new(vid)
				@parser = EuroParser
				$logger.debug("Downloader: For Euro mode...")
			else
				@map = NinjinCountryMap.new(vid)
				@parser = NinjinParser
				$logger.debug("Downloader: For Ninjin mode...")
			end
			@result = true
			@config = Config.new
			@user, @pass = '', ''
			Dir.mkdir(Software::LogDir) unless File.exist?(Software::LogDir)
	
			if @map['active']
				case @map['country']
				when 'C'
					@user = @config['user_id_c']
					@pass = @config['password_c']
				when 'F'
					@user = @config['user_id_f']
					@pass = @config['password_f']
				when 'O'
					@user = @config['user_id_o']
					@pass = @config['password_o']
				end
			end
		end

		def self.update(vid)
			new(vid).update
		end

		def update
			$logger.debug("Downloader: uri => #{@map['uri4get']}, dir => #{Software::LogDir + @map['vid']}")
			if File.exist?(Software::LogDir + @map['vid'])
				update_part
			else
				Dir.mkdir(Software::LogDir + @map['vid'])
				update_all
			end
		end

		#private
	
		# ʤǤ update Ǥ update
		def update_part
			$logger.debug("Downloader: Updating now...")
			w = httpget(Software::LogDir + "#{@map['vid']}/tmp", false, true)
			if w
				add_logs = []
				ids = []

				hash = @parser::parse(@map['vid'], 'tmp', false)
				hash['content'].each {|msg| ids << msg.id }
				ids = ids - last_ids
				ids.each {|id|
					hash['content'].each {|msg| 
						if msg.id == id
							add_logs << msg.to_s
							hash['content'].reject! {|x| x == msg }
						end
					}
				}

				today = page_day(NKF.nkf('-ex', File.read(Software::LogDir + @map['vid'] + '/tmp')))

				$logger.debug("Downloader: Partical?: #{Software::LogDir + @map['vid'] + '/' + today + '.dat'}, #{ids.size}")
				if add_logs == []
					$logger.debug("Downloader: Nothing? add_logs == [] !!")
				elsif File.exist?(Software::LogDir + "#{@map['vid']}/#{today}.dat") && \
					ids.size < @map['tmppage_voicecnt']
					# partical
					$logger.debug("Downloader: #{today} log partical updated...")
					lf = Dir.glob(Software::LogDir + "#{@map['vid']}/*.dat").sort.last
					add_logs = add_logs.join("\n") + "\n"
					File.open(lf, 'a') {|fh| fh.write add_logs }
				else
					$logger.debug("Downloader: #{today} log partical update failed. Turn to download all...")
					update_all
				end

				true
			else
				$logger.debug("Downloader: Not Updating")
				@result = false
				false
			end
		end
	
		#  update
		def update_all
			$logger.debug("Downloader: Downloading all now...")
			w = httpget(Software::LogDir + "#{@map['vid']}/tmp", @map['mes_all'], false)
			if w
				tmp = NKF.nkf('-ex', File.read(Software::LogDir + "#{@map['vid']}/tmp"))
				today = page_day(tmp)
				uris = page_uris(tmp)
				l = Logs.new(@map['vid'], @map['logkey'])
				uris.each {|ary| update1day(ary[1], ary[0]) }
				l.reload
				$logger.debug("Downloader: Downloading all done. Parsing today's log(#{today}).")
				if today == 'end'
					File.unlink(Software::LogDir + "#{@map['vid']}/tmp")
				else
					if today != @map['party']
						today = 'day' + today if today !~ /^day/
					end
					File.rename(Software::LogDir + "#{@map['vid']}/tmp",
						Software::LogDir + "#{@map['vid']}/#{today}")
					@parser::parse(@map['vid'], today)
					File.unlink(Software::LogDir + "#{@map['vid']}/#{today}")
				end
				true
			else
				@result = false
				false
			end
		end

		#  update
		def update1day(uri, date)
			unless File.exist?(Software::LogDir + "#{@map['vid']}/.#{date}.done")
				$logger.debug("Downloader: Updating #{@map['vid']}/#{date}...(#{uri})")
				httpget(Software::LogDir + "#{@map['vid']}/#{date}", uri)
				@parser::parse(@map['vid'], date)
				File.touch(Software::LogDir + "#{@map['vid']}/.#{date}.done")
				File.unlink(Software::LogDir + "#{@map['vid']}/#{date}")
			end
		end

		# ڡ
		def page_day(str)
			today = '0'
			str.gsub(@map['regex_today_cut'], '').gsub(@map['regex_today']) {
				td = $1.dup
				case td
				when @map['regex_today_days'] then today = 'day' + $1.dup
				when @map['regex_today_epi'] then today = 'party'
				when @map['regex_today_end'] then today = 'end'
				end
			}
			$logger.debug("Downloader#page_day: today => #{today}")
			today
		end

		# ڡ󥯤Ƥڡ URI Array
		def page_uris(str)
			ret = str.gsub(@map['regex_pageuris_head_end_0'], '').gsub(@map['regex_pageuris_head_end_1'], '')
			ret = ret.gsub(@map['regex_pageuris_foot_start'], '').split("\n")[0]
	
			uris = []
			ret.gsub(@map['regex_pageuris_uris']) {
				arg = $1.dup
				case arg
				when @map['regex_pageuris_ready'] then uris << ['day0', arg]
				when @map['regex_pageuris_party'] then uris << ['party', arg]
				when @map['regex_pageuris_days']
					if @map['pageuris_days_succ']
						uris << ['day' + $1.succ, arg]
					else
						uris << ['day' + $1.dup, arg]
					end
				end
			}
			$logger.debug("Downloader#page_uris: uris => #{uris.flatten.join(', ')}")
			uris
		end

		# ǿ 10 ܤޤǤ Message  ID ֤(ո٤ǽ)
		def last_ids
			logs = Dir.glob(Software::LogDir + "#{@map['vid']}/*.dat").sort
			lary = []
	
			IO.readlines(logs.last).reverse_each {|line|
				lary << line.split('<>')[0]
			}
			if lary.size < 10
				IO.readlines(logs[logs.size - 2]).reverse_each {|line|
					lary << line.split('<>')[0]
				}
			end
	
			lary[0..9]
		end
	
		def httpget(fp, path, chk = false)
			if chk
				if need_update?(fp + '.chk')
					_httpget(@map['vid'], path, fp)
				else
					false
				end
			else
				_httpget(@map['vid'], path, fp)
			end
		end

		# Downloader#httpget (Action ֥Ȼ)
		def _httpget(vid, path, fp)
			act = Action.new(vid, @user, @pass)
			act.get(fp, path)
		end

		# RSS  Head Last-Modified Ӥ׹ɤ֤
		def need_update?(chkfile)
			File.touch(chkfile) unless File.exist?(chkfile)

			act = Action.new(@map['vid'], @user, @pass)
			res = act.head
			if res['last-modified']
				last = Time.parse(res['last-modified']).to_s
			else  # Ϥʤ HEAD  last-modified ֤ʤΤ...
				last = 'unknown'
			end
			past = File.read(chkfile)

			$logger.debug("Downloader: Local file time is #{past}.")
			$logger.debug("Downloader: Remote file time is #{last}.")

			if last == past
				false
			else
				File.open(chkfile, 'w') {|fh| fh.write last } unless last == 'unknown'
				true
			end
		end
	end

	#  HTML  ZeX Dat Ѵ parser
	class ParserBase
		def initialize(vid, filename, compat)
			@fn = filename
			@compat = compat
		end

		# Ū˳ϤΥ饹᥽åɤǸƤӽФ
		# Ѥܥ᥽åɤ
		#  _parse, _parse_extra, _parse_votes, update_time, epi_rolltable  5ġ
		# Ѥ륯饹
		# Clip  ExtraParser  2ġ
		#
		# (String, String, true/false)
		#  * filename  File.basename Υե̾
		#  * compat  dat Ŭǡ֤ݤ(backward compatibility)
		def self.parse(vid, filename, compat = true)
			$logger.debug("Parser: Parsing now... (#{vid}, #{filename})")
			new(vid, filename, compat).parse
		end

		#  HTML ե parse
		# ͤ Content => ᥤȯ, extra => إå Hash
		def parse
			d = false
			d = @fn.sub(/^day/, '')
			dir = Software::LogDir + @map['vid'] + '/'
			log = NKF.nkf('-ex', File.read(dir + @fn))
			announce = false

			unless @fn == 'tmp'
				File.dump(dir + 'name', village_name(log)) unless File.exist?(dir + 'name')
				File.dump(dir + 'roll', epi_rolltable(log)) if @fn == 'party'
				if @need_idtable  # (͹Ĵ)
					File.dump(dir + 'idtable', _generate_idtable(d, log).to_s)
					File.touch(dir + '.idtable.done') if d == '2'
				end
				announce = _parse_extra(d, log)
			end

			log = @clip::header(d, @clip::footer(log))
			hash = {'content' => _parse(log), 'extra' => announce}

			if hash['extra'] && hash['extra'].has_key?('epilogue')
				File.touch(dir + '.party')
			end

			if @compat
				hash = convert2data(hash)
				{'.dat' => hash['content'], '.ann' => hash['extra']}.each {|k, v|
					File.dump("#{dir}/#{@fn}#{k}", v)
				}
			else
				hash
			end
		end

		# _parse Ǽ Object Hash  dat ˳Ǽ䤹 String Hash 
		def convert2data(hash)
			h = {'content' => '', 'extra' => ''}
			hash['content'].each {|msg| h['content'] << msg.to_s + "\n" }

			if hash['extra'].has_key?('vote') 
				if hash['extra'].has_key?('random')
					h['extra'] << 'ɼ: ' + hash['extra']['random'].join(' ') + "\n"
				end
				hash['extra']['vote'][1].each {|to, from|
					h['extra'] << "#{to}(#{from.size}): #{from.join(' ')}\n"
				}
				if hash['extra']['vote'][0] != ''
					h['extra'] << "\n"
					h['extra'] << '' + hash['extra']['vote'][0] + "\n"
				end
			end
			a = [['newtype', ''], ['medium_result', ''], ['fortune', ''], 
				 ['fortune_result', ''], ['hunter', ''],['attack', '']]
			a.each {|ary|
					if hash['extra'].has_key?(ary[0]) && hash['extra'][ary[0]] != []
						h['extra'] << ary[1] + hash['extra'][ary[0]].join(' ') + "\n"
					end
			}
			h['extra'] << "\n"
			%w(update_time rolltable chance epilogue).each {|key|
				h['extra'] << hash['extra'][key].to_s + "\n" if hash['extra'].has_key?(key)
			}
			h
		end

		# "¼"ʸ֤
		def village_name(log)
			vname = ''
			log.gsub(/<title>(.*)<\/title>/) {
				ary = $1.split(' ')
				vname = ary[1] + ' ' + ary[2]
			}
			vname
		end

		# 1ܤ򿦿Ϳ parsexX, .. ʸ֤
		def rolltable(announce)
			rtable = []
			announce.scan(/(.*?)(\d)(̾|)/) {|sary|
				rtable << sary[0] + 'x' + sary[1] + ' '
			}
			rtable.join(' ') + "\n"
		end

		# ̾饤˾褻Υإåȥեåν
		class Clip
			def initialize
			end

			def self.header(d, log)
				new().header(d, log)
			end

			def self.footer(log)
				new().footer(log)
			end

			def header(d, log)
				if d == @prolog_date
					ret = log.gsub(@prolog_pages_header, '')
				else
					ret = log.gsub(@short_pages_header, '')
					ret =~ @normal_days_extra
					ret.sub(@normal_days_extra, '')
				end
			end

			def footer(log)
				ret = log.gsub(@general_footer, '')
				if ret =~ (@inprogress_order_mark)  # in progress
					ret.gsub(@inprogress_order, '')
				else
					ret.gsub(@general_order, '')
				end
			end
		end

		# ߼η׻ɽ
		class ExecuteCalculation
			# Ϳȥߡʤˡ߼ܤʸ֤
			def self.calculate(num, gerd = false)
				return('') if num.to_i == 0
				ret = num.to_s + ': '
				ary = [num.to_s]
			
				i = num.to_i
				if gerd
					i -= 1
					ary << i.to_s
				end
			
				while i > 4
					i -= 2
					ary << i.to_s
				end

				if gerd
					ret + ary.join('-') + '-EP x' + (ary.size - 1).to_s
				else
					ret + ary.join('-') + '-EP x' + ary.size.to_s
				end
			end
		end
	end

	class EuroParser < ParserBase
		def initialize(vid, filename, compat)
			@map = EuroCountryMap.new(vid)
			@clip = EuroClip
			@extraparser = EuroExtraParser
			@need_idtable = true
			super
		end

		# ᥤ󥳥ƥ(ȯʬ) parse
		def _parse(log)
			msgs = []
			scanner = StringScanner.new(log)
			loop {
				scanned = false

				[['msg', %r{<div class="message">(.*?)</table>(</div>)+}m],
				 ['act', %r{<div class="act">(.*?)</div>}m],
				 ['sys', %r{<div class="think">(.*?)</div>}],
				 ['sys', %r{<div class="system">(.*?)</div>}],
				 ['end', %r{</td></tr><tr>.*}]].each {|ary|
					s = scanner.scan(ary[1])
					if s
						scanned = true
						unless ary[0] == 'end'
							msgs << __send__("_parse_#{ary[0]}", scanner[1])
						end
					end
				}

				unless scanned
					$logger.debug("Parse(StrScan): Unknown Pattern (#{scanner.peep(60)})")
					raise('unknown pattern')
				end
				break if scanner.eos?
			}
			return(msgs)
		end

		def _parse_msg(block)
			msg = Message.new
			block.gsub(/<a name="(.*?)"><\/a>/) {
				msg.id = $1.dup
				msg.type = $1.gsub(/\d*/, '')
			}
			block.gsub(/<a class="name" href=".\/index.rb\?vil_id=\d*&id=(\d*)".*?>(.*?)<\/a>/) {
				msg.icon = $1.dup
				msg.person = $2.split(' ')[1]
			}
			unless msg.person
				block.gsub(/<span class="prv">(.*?)<\/span>/) { msg.person = $1.split(' ')[1] }
			end
			unless msg.person  # ˵Ѽ
				block.gsub(/<a href=".\/matome.rb\?.*?" .*?>(.*?)<\/a>/) { msg.person = $1.dup }
			end
			block.gsub(/<div class="mes_.*?_body1">(.*?)<\/div>/) {
				msg.content = $1.gsub(/(<a.*?>|<\/a>)/, '')
			}
			block.gsub(/<span class="time">(.*?)<\/span>/) { msg.time = $1.dup }
			msg.id = 'say0' unless msg.id
			msg.type = 'say' unless msg.type
			msg.icon = '1' unless msg.icon
			msg
		end

		def _parse_act(blockbase)
			msg = Message.new
			msg.id, msg.type, msg.icon = 'act', 'act' , ''
			block = blockbase.gsub(/(<a.*?>|<\/a>)/, '')
			block.gsub(/<span class="acttime">(.*?)<\/span>/) { msg.time = ($1).split(' ')[1] }
			block.gsub(/(.*)<span class="acttime">/) {
				msg.content = $1.dup
				msg.content.gsub(/^(.*?)ϡ.*/) { msg.person = $1.split(' ')[1] }
			}
			msg
		end

		def _parse_sys(block)
			msg = Message.new
			msg.id, msg.type, msg.icon, msg.time = 'sys', 'sys' , '', nil
			msg.content = block
			msg.content.gsub(/(.*?)(|)/) { msg.person = $1.split(' ')[1] }
			msg
		end

		# ʥ󥹡ȥʬ parser / ̥饹̥᥽åɻ
		def _parse_extra(d, log)
			block = ''
			alive = 0
			if d =~ /[23]/
				idtable = IDTable.new(@map['vid'])
				idtable.load
				alive = idtable.keys.size.to_i
			end
			ann_contents = []
			extras = {}

			top = %Q(</td>\n</tr>\n</table></td><td valign="top">)
			log.gsub(%r{.*#{top}(.*?)<div class="message">}) { block = $1.dup }
			block.gsub(%r{<div class="(think|kill|system|win)">(.*?)</div>}) {
				x = $2.dup
				ann_contents << x unless x =~ /(|路)/
			}

			case d
			when '1' then extras['update_time'] = ''  # ϱ³Ūɽʤ
			when '2'
				extras['rolltable'] = rolltable(ann_contents[0])
				extras['chance'] = ExecuteCalculation.calculate(alive, true)
			else
				exparser = @extraparser.new
				ann_contents.each {|x| exparser.parse(x) }
				extras = exparser.data
				v = [exparser.data['vote_result'].to_s]
				x, tmp_alive = _parse_votes(exparser.data['vote'])
				alive = tmp_alive - 1 unless d == '3'  # 跺 -1
				alive = alive - _calc_attack_num(exparser.data['attack'])  # 跺ब 1
				v << x
				extras['vote'] = v
				unless extras.has_key?('epilogue')
					extras['chance'] = ExecuteCalculation.calculate(alive) if d
				end
			end

			extras
		end

		# Կη׻
		def _calc_attack_num(data)
			return(0) unless data
			ary = data[0].split(' ')
			if ary.size == 2
				2
			elsif ary[0] == 'ʤ'
				0
			else
				1
			end
		end

		# 跺ԤФɼνסɽ
		# ɼʸι Array
		# ͤ [String, Hash] => [跺, {'ɼ' => ['ɼ', ..]}]
		# Ǥͤ Hash => {'ɼ' => ['ɼ', ..]}  parse_extra 
		def _parse_votes(ary)
			alive = 0
			votes = {}
			ary.each {|vote|
				vary = vote.split(' ')
				from = vary[1].gsub(/<\/td>.*/ , '')
				to = vary[2].gsub(/<\/td>.*/ , '')
				votes[to] = [] unless votes.has_key?(to)
				votes[to] << from
				alive += 1
			}
			[votes, alive]
		end

		# ԥɽɽ parsetable ʸ֤
		def epi_rolltable(log)
			ret = "<tr><td>#{['', 'ID', '', 'ǽ'].join('</td><td>')}</td></tr>\n"
			tbl = log.gsub(%r{.*<table class="member">(.*?)</table>.*}m, '\1')
			state = '¸'
			tbl.each {|line|
				state = $1.dup if line =~ %r(<h3>(.*?)\s*.*</h3>)
				if line =~ %r(<a href=.*?>.*? (.*?)</a><br>.*?>(.*?)</a><br>(.*)$)
					ret << "<tr><td>#{[$1, $2, state, $3].join('</td><td>')}</td></tr>\n"
				end
			}
			"<table>\n" + ret + "</table>\n"
		end

		#  person id  person б༡ʤᡢΥơ֥(Hash)֤
		def _generate_idtable(day, log)
			table = IDTable.new(@map['vid'])
			if File.exist?(Software::LogDir + @map['vid'] + '/.idtable.done')
				table.load
			else
				if day == '1' || day == '2'
					log.gsub(/<table class="member">(.*?)<\/table>/m) {
						$1.each {|line|
							if line =~ /<a href=".\/.*?\?vil_id=\d*&id=(\d*)".*?">(.*?)<\/a>/
								table[$1.dup] = $2.split[1]
							end
						}
					}
				end
			end

			table
		end

		class EuroClip < ParserBase::Clip
			def initialize
				@prolog_pages_header = %r{.*ä礤ϤΤǤ</div>}m
				@short_pages_header = %r{.*&log=all">٤ɽ</a></div></div>}m
				@normal_days_extra = %r{.*?(<div class="(system|kill|think|win)">.*?</div>)+}m
				@general_footer = %r{</td></tr><tr><td></td><td valign="top"><a name="form"></a></td></tr>.*}m
				@inprogress_order_mark = %r{<div class="message"><div class="order">}
				@inprogress_order = %r{<div class="message"><div class="order">.*}m
				@general_order = @general_footer
				@prolog_date = '1'
			end
		end

		# Header ʬ(announce, extra) parser
		class EuroExtraParser
			attr_reader :data

			def initialize
				@data = {'vote' => []}
			end

			def parse(e)
				case e
				when // then type = 'newtype'
				when // then type = 'random'
				when /ɼ/ then type = 'vote'
				when /跺/ then type = 'vote_result'
				when // then type = 'hunter'
				when // then type = 'fortune'
				when /Τ褦Ǥ/ then type = 'fortune_result'
				when /ä褦Ǥ/ then type = 'medium_result'
				when /̵|/ then type = 'attack'
				when // then type = 'epilogue'
				end
				if type
					@data[type] = [] unless @data.has_key?(type)
					@data[type] << __send__("parse_#{type}", e)
					@data[type].flatten!
				else
					$logger.debug("EuroExtraParse: ignore extra, #{e}")
				end
			end

			def parse_newtype(e)
				e.split(' ')[1]
			end

			def parse_random(e)
				e.gsub(/.*? (.*?).*? (.*?)ɼ.*/) {
					"#{$1.dup}(#{$2.dup})"
				}
			end

			def parse_vote(e)
				e.gsub(%r{(<table><tr>|</tr></table>)}, '').split('</tr><tr>')
			end

			def parse_vote_result(e)
				e.split(' ')[1]
			end

			def parse_hunter(e)
				ary = e.split(' ')
				ary[4] + ' by ' + ary[1]
			end
			alias :parse_fortune :parse_hunter

			def parse_fortune_result(e)
				ary = e.split(' ')
				ary[1] + '(' + ary[3].gsub(/(|ä)褦Ǥ/, '') + ')'
			end
			alias :parse_medium_result :parse_fortune_result

			def parse_attack(e)
				if e =~ /̵/
					ary = e.split(' ')
					if e =~ /  / # ham melting
						ary[1] + ' ' + ary[4]
					else
						ary[1]
					end
				else
					'ʤ'
				end
			end

			def parse_epilogue(e)
				e
			end
		end
	end

	class IDTable < Hash
		attr_reader :r

		def initialize(vid)
			@vid = vid
			@r = {}
			super
		end

		def to_s
			ret = ''
			self.each {|k, v| ret << "#{k}:#{v}\n"}
			ret
		end

		def load
			fp = Software::LogDir + @vid + '/idtable'
			if File.exist?(fp)
				File.read(fp).each {|line|
					ary = line.chomp.split(':')
					self[ary[0]] = ary[1]
					@r[ary[1]] = ary[0]
				}
			else
				Library::Ninjin::Persons.each {|k, v|
					self[k] = v
					@r[v] = k
				}
			end
		end
	end

	class NinjinParser < ParserBase
		def initialize(vid, filename, compat)
			@map = NinjinCountryMap.new(vid)
			@clip = NinjinClip
			@extraparser = NinjinExtraParser
			super
		end

		# ᥤ󥳥ƥ(ȯʬ) parse
		def _parse(log)
			msgs = []
			log.split("</table>\n\n</div>\n").each {|block|
				msg = Message.new
				block.gsub(/^<a name="mes(\d+)">(.*?)<\/a>/) {
					msg.id = $1.dup
					msg.person = $2.split(' ')[1]
				}
				block.gsub(/#{@map['image']}face(\d\d)\.jpg/) { msg.icon = $1.dup }
				block.gsub(/#{@map['image']}(.*?)00\.jpg/) { msg.type = $1.dup }
				block.gsub(/<div class="mes_.*?_body1">(.*)<\/div>$/) { msg.content = $1.dup }
				block.gsub(/<span class="time">(.*)<\/span>/) {
					t = $1.gsub(/(\s|ʬ)/, '').gsub(//, ':')
					if t.index("")
						t.concat(" AM").gsub!(//, '')
					else
						t.concat(" PM").gsub!(//, '')
					end
					msg.time = Time.parse(t).strftime("%H:%M")
				}
				msgs << _parse_sys(block) if block.index(%Q(div class="announce"))

				msgs << msg
			}
			return(msgs)
		end

		# Xܡ
		def _parse_sys(block)
			msg = Message.new
			msg.id, msg.type, msg.icon, msg.time, msg.person = 'sys', 'sys' , '', nil, nil
			block.gsub(/<div class="announce">\n(.*?)\n<\/div>\n<\/div>/m){ msg.content = $1.dup }
			msg
		end

		# ʥ󥹡ȥʬ parser / ̥饹̥᥽åɻ
		def _parse_extra(d, log)
			alive = 0
			ann_contents = []
			extras = {}
	
			log.gsub(%r{<div class="(ann|announce|extra)">\n(.*)\n</div>}) {
				x = $2.dup
				ann_contents << x unless x =~ /(¸|路)/
				x.gsub(/(\d+) ̾/) { alive = $1.to_s } if x =~ /¸/
			}

			case d
			when '0' then extras['update_time'] = update_time(log)
			when '1'
				rtable = rolltable(ann_contents[1])
				rtable.scan(/(\d)/) {|num| alive += num.to_s.to_i }
				extras['rolltable'] = rtable
				extras['chance'] = ExecuteCalculation.calculate(alive, true)
			else
				exparser = @extraparser.new
				ann_contents.each {|x| exparser.parse(x) }
				extras = exparser.data
				extras['vote'] = _parse_votes(exparser.data['vote'])
				extras['chance'] = ExecuteCalculation.calculate(alive) if d
			end

			extras
		end

		# 跺ԤФɼνסɽ
		# ɼʸι Array
		# ͤ [String, Hash] => [跺, {'ɼ' => ['ɼ', ..]}]
		def _parse_votes(ary)
			convict, votes = '', {}

			ary.delete('')
			ary.each {|vote|
				vary = vote.split(' ')
				if vote =~ /跺/
					convict = vary[1]
				else
					votes[vary[4]] = [] unless votes.has_key?(vary[4])
					votes[vary[4]] << vary[1]
				end
			}

			[convict, votes]
		end

		# 1ܤΥ鹹֤ʸ֤
		def update_time(log)
			log.gsub(/.*<strong>\d+\/\d+(.*)<\/strong>.*/m, '\1' + "\n")
		end

		# ԥɽɽ parsetable ʸ֤
		def epi_rolltable(log)
			reg = /<div class="announce">\n(ŷ  .*?ˡ˴¼ͤä.*?)$/
			ret = "<tr><td>#{['', 'ID', '', 'ǽ'].join('</td><td>')}</td></tr>\n"
			log.gsub(reg) {
				$1.split('<br />').each {|r|
					r =~ /(.*?) (.*?) (.*?)ˡ(.*?)(.*?)ä/
					ret << "<tr><td>#{[$2, $3, $4, $5].join('</td><td>')}</td></tr>\n"
				}
			}
			"<table>\n" + ret + "</table>\n"
		end

		class NinjinClip < ParserBase::Clip
			def initialize
				@prolog_pages_header = %r{.*?Ȥˤʤä<br />\n</div>\n</div>\n\n}m
				@short_pages_header = %r{.*<a href="index.rb\?vid=\d+&amp;mes=all">ɽ</a>\n}m
				@normal_days_extra = %r{.*<div class="announce">\n.*\n</div>\n</div>\n}m
				@general_footer = /<div class="footer">.*/m
				@inprogress_order_mark = %r{<div class="order"}
				@inprogress_order = /\n{2,3}<div class="message">\n<div class="order">\n.*/m
				@general_order = /<a href=.*\n\n<\/div>\n/
				@prolog_date = '0'
			end
		end

		# Header ʬ(announce, extra) parser
		class NinjinExtraParser
			attr_reader :data

			def initialize
				@data = {}
				@data = {'vote' => []}
			end

			def parse(e)
				case e
				when // then type = 'newtype'
				when /ɼ/ then type = 'vote'
				when // then type = 'hunter'
				when // then type = 'fortune'
				when /̵|/ then type = 'attack'
				when /Ƥοϵ|̤γʪ|/ then type = 'epilogue'
				end
				if type
					@data[type] = [] unless @data.has_key?(type)
					@data[type] << __send__("parse_#{type}", e)
					@data[type].flatten!
				end
			end

			def parse_newtype(e)
				e.split(' ')[1]
			end

			def parse_vote(e)
				e.split('<br />')
			end

			def parse_hunter(e)
				ary = e.split(' ')
				ary[3] + ' by ' + ary[1]
			end

			alias :parse_fortune :parse_hunter

			def parse_attack(e)
				if e =~ /̵/
					ary = e.split(' ')
					if e =~ /  / # ham melting
						ary[1] + ' ' + ary[4]
					else
						ary[1]
					end
				else
					'ʤ'
				end
			end

			def parse_epilogue(e)
				e.gsub(/<br \/>/, '')
			end
		end
	end

	# Ƥȯλ key ˤ Hash, value  Message ֥Ȥ Array
	class Logs < Hash
		def initialize(vid, key, type = 'all')
			$logger.debug("Logs.new: for #{vid} (key=#{key})")
			@vid = vid
			@key = key
			@type = type
			@fp = Software::LogDir + "#{@vid}/log"
			@max_day = 15
			super()
	
			if File.exist?(@fp)
				load
				# backword compatibility
				reload if self == {}
				reload unless File.exist?(@fp + '.keys')
			else
				create
			end
		end
	
		def create
			self.clear
			0.upto(@max_day) {|day|
				if File.exist?(Software::LogDir + "#{@vid}/.day#{day}.done")
					IO.readlines(Software::LogDir + "#{@vid}/day#{day}.dat").each {|line|
						msg = Message.new(line.chomp.split('<>'))
						(self["#{day}:#{msg[@key]}"] ||= []).push msg
					}
				end
			}
			dump
			type_reject!
		end
		alias :reload :create
	

		private

		def load
			File.open(@fp, 'rb') {|fh| self.update(Marshal.load(fh.read)) }
			type_reject!
		end
	
		def dump
			File.open(@fp, 'wb') {|fh| fh.write Marshal.dump(self) }
			link_keys = Hash.new
			self.keys.each {|key|
				self[key].each {|msg| (link_keys[key] ||= []).push msg.type }
			}
			File.open(@fp + '.keys', 'wb') {|fh| fh.write Marshal.dump(link_keys) }
		end

		def type_reject!
			$logger.debug("Logs#type_reject! #{@type}")
			case @type
			when 'white'
				self.each {|key, value|
					value.reject! {|msg| msg.type =~ /(think|whisper|groan)/o }
				}
			when 'whisper'
				self.each {|key, value|
					value.reject! {|msg| msg.type =~ /(think|groan)/o }
				}
			end
		end
	end

	# ҤȤĤȯб륪֥
	class Message < Hash
		@@keys = %w(id icon type time content person)
		@@keys.each {|key|
			define_method(key) { self[key] }
			define_method("#{key}=") {|arg| self[key] = arg }
		}
	
		def initialize(ary = false)  # [id, icon, type, time, content, person]
			if ary
				@@keys.each_with_index {|key, i| self[key] = ary[i] }
			end
		end

		def to_s
			if self['content'].index('<div class="loud">')
				self['content'].gsub!(/<div class="loud">/, '<span class="loud">')
				self['content'] = self['content'] + '</span>'
			end

			ary = []
			@@keys.each {|key| ary << self[key] }
			ary.join('<>')
		end

		def std?
			@std ||= /(say|think|whisper|groan)/
		end
	
		# Message Υ᥿
		def sign(day, voice_cnt)
			typeid = ''
			if self['id'] =~ /^\d+$/  # Ninjin Type ID
				case day
				when 'party' then d = 'Party '
				else
					d = day + 'D'
				end
				time = d + self['time'] + '['
				time << Time.at(self['id'].to_i).strftime("%S") + ']'
			else  # other
				time = self['time']
				typeid = '[' + day + ':' + mark(self['type'])
				typeid << self['id'].delete(self['type']) + ']'
			end

			case self['type']
			when 'act' then sign = %Q((#{self['time']}))
			when 'sys' then sign = ''
			else
				sign =  %Q(#{self['person']} #{time})
				sign << %Q((#{voice_cnt})#{typeid})
				sign
			end

			sign
		end

		def mark(type)
			h = {'say' => '', 'whisper' => '*', 'groan' => '+', \
				 'think' => '-', 'act' => '', 'sys' => ''}
			h[type]
		end
	end

	# View  Header ʬ
	module Header
		module_function

		def set(vid, day, type)
			@@vid, @@day, @@type = vid, day, type
			@@day = 'day' + @@day unless @@day == 'party'
		end

		def announce
			if File.exist?(Software::LogDir + @@vid + '/' + @@day + '.ann')
				type_reject(File.read(Software::LogDir + @@vid + '/' + @@day + '.ann').gsub(/\n/, "<br>\n"))
			end
		end

		def type_reject(announce)
			case @@type
			when 'all' then announce
			else  # 'white' & 'whisper'
				x = announce.gsub(/^(||||).*\n/, '')
				x.gsub(/<br>/, '').strip.gsub(/\n/, '<br>')
			end
		end

		def title
			if File.exist?(Software::LogDir + @@vid + '/name')
				day = @@day.gsub(/day/, '')
				title = File.read(Software::LogDir + @@vid + '/name') + "(#{day.capitalize})"
			end
		end
	end

	# ǽ֤᥽åɤΤߤΥ⥸塼
	module LastDaySearch
		def last_day_search(vid)
			dat = Dir.glob(Software::LogDir + "#{vid}/*.dat").sort.last
			File.basename(dat, '.dat').gsub(/day/, '')
		end
	end
	

	# ᥤ̤Υɥ˥塼
	class MainMenu
		def initialize(vid, day, sword, emulation, vcnt, tcnt)
			@vid, @d, @s, @emu, @vcnt, @tcnt = vid, day, sword, emulation, vcnt, tcnt
			@config = Config.new
			@idtable = IDTable.new(vid)
			@idtable.load
		end

		# Ū˳ϤΥ饹᥽åɤǸƤӽФ
		#
		# (String, String, RealTimeEmulate, Hash, Array)
		#  * vcnt(Hash) => ʪ ID  ȯ Hash
		#  * tcnt(Hash) => ȯ  ȯ Hash
		#  * types(Array) => ɽǤĻϿΥפ Array
		#    types[0] ɬɽƤĻϿפˤ롣
		#
		def self.generate(vid, day, sword, emulation, vcnt, types, tcnt)
			new(vid, day, sword, emulation, vcnt, tcnt).generate(types)
		end

		def generate(types)
			$logger.debug("MainMenu#generate: #{types.join(', ')}")
			persons = []
			@vcnt.each {|k, v| persons << k }
			@t = types[0]

			d = dateselector(@d, @vid, @emu)
			if File.exist?(Software::LogDir + @vid + '/party.dat') && @emu.emulating? == false
				d << list("#{img('world')} #{link("http://wolfbbs.jp/#{@vid}%C2%BC.html", 'wolfbbs')}")
			end

			if @emu.emulating?
				d << list(link("?mode=prt_plus&vid=#{@vid}", '1ʤ'))
				d << list(link("?mode=prt_end&vid=#{@vid}", 'P/RTｪλ'))
			else
				d << list("#{img('chart_curve')} #{link("./?mode=status&vid=#{@vid}", 'ȯ')}")
			end
	
			t = typeselector(types) unless @emu.emulating?
	
			r = ''
			[d, t].each {|ary|
				if ary && ary != []
					r << "<ul>\n" + ary.join("\n") + "</ul>\n"
				end
			}
			tary = typefilter()
			r << tary.join("\n") if tary.size > 3
			r << personselector(persons).join("\n")
			@s = '' unless @s  # for search.part
			r << ERB.new(Template['part/mainmenu']).result(binding)
			r
		end

		# target Υ󥯤 content Ȥ٥֤
		def link(target, content)
			%Q(<a href="#{target}">#{content}</a>)
		end

		# li ǤǰϤʸ֤
		def list(content)
			'<li>' + content + '</li>'
		end

		def img(fn)
			%Q(<img src="#{Software::ImageDir}#{fn}.png">)
		end

		# ȯ̥ե륿
		def typefilter
			ary = [%Q(<form name="types">\n<ul>)]
			n = 0
			[['say', '̾ȯ', ''],
			 ['think', 'Ȥ', ''],
			 ['whisper', '䤭', ''],
			 ['groan', '᤭', '']].each {|type|
				if @tcnt.has_key?(type[0])
					n += 1
					if @config['compact_mode']
						x = typecheckbox(type[0], @tcnt[type[0]], type[2])
						x = '<li>' + x if n == 1
						x = x + '</li>' if n == 2
					else
						x = typecheckbox(type[0], @tcnt[type[0]], type[1])
						x = '<li>' + x + '</li>'
					end
					ary << x
					n = 0 if n == 2
				end
			}
			ary << %Q(</ul>\n</form>)
			ary
		end

		# ȯץե륿Ǥ٥դ
		def typecheckbox(tid, tcnt, label)
			ret = ''
			ret << %Q(<input type="checkbox" name="#{tid}" value="#{tcnt}" onclick="thidden('#{tid}');" )
			ret << %Q(checked="checked"><a onclick="tcheckwithinline('#{tid}');">#{label}</a>)
			ret
		end

		# ʪե륿
		def personselector(pids)
			ary = [%Q(<form name="persons">\n<ul>)]
			pids.reject! {|x| x == '99' }  # Ninjin ''
			pids.sort! {|a, b| a.to_i <=> b.to_i }

			n = 0
			pids.each {|pid|
				n += 1
				name = @idtable[pid]
				if @config['compact_mode']
					x = personcheckbox(pid, @vcnt[pid], Library::Rolls[name][1])
					x = '<li>' + x if n == 1
					x = x + '</li>' if n == 2
				else
					x = personcheckbox(pid, @vcnt[pid], name)
					x = '<li>' + x + '</li>'
				end
				ary << x
				n = 0 if n == 2
			}
			ary << %Q(</ul>\n</form>)
			ary
		end

		# ʪե륿Ǥ٥դ
		def personcheckbox(pid, pcnt, label)
			ret = ''
			ret << %Q(<input type="checkbox" name="#{pid}" value="#{pcnt}" onclick="hidden('#{pid}');" )
			ret << %Q(checked="checked"><a onclick="checkwithinline('#{pid}');">#{label}</a>)
			ret
		end

		# 0ܡѡƥޤǤؤΥ󥯥˥塼
		def dateselector(day, vid, emulation)
			ary = []

			if File.exist?(Software::LogDir + vid + '/day0.dat')
				ary << menulink(day, '0', './?day=0', 'ץ')
			end

			Dir.glob(Software::LogDir + vid + '/day*.dat').sort.each {|dat|
				if dat =~ /^#{Software::LogDir + vid}\/day(\d)\.dat$/
					j = $1.dup
					unless j == '0'
						if emulation.emulating? && emulation.state.include?(j) == false
						else
							ary << "<li>\n" if j == '1' or j == '5' or j == '9'
							ary << menulink(day, j, "./?day=#{j}", "#{j}D", false) + ' '
							ary << '</li>' if j == '4' or j == '8'
						end
					end
				end
			}
			ary << '</li>' if ary.last != '</li>'

			if File.exist?(Software::LogDir + vid + '/party.dat') && emulation.emulating? == false
				ary << menulink(day, 'party', "./?day=party", 'ѡƥ')
			end

			ary
		end

		# ɽΥ(¼ϵ)˥塼
		def typeselector(types)
			return([]) if types.size == 1
			ary = []
			ary << '<li class="type">['
			ary << menulink(@t, 'white', './?type=white', '¼', false) + ' | '
			if types.include?('whisper')
				ary << menulink(@t, 'whisper', './?type=whisper', 'ϵ', false) + ' | '
			end
			ary << menulink(@t, 'all', './?type=all', '', false)
			ary << "]</li>\n"
			ary
		end
	
		# ñ CGI ѥ᥿ʤ a Ǥ type  day ä a Ǥ֤
		def savelink(link)
			link.gsub(%r{<a href="\./\?(.*?)=(.*?)(#.*)*">(.*)</a>}) {
				case $1
				when 'day' then base = %Q(type=#{@t})
				when 'type' then base = %Q(day=#{@d})
				end
				%Q(<a href="./?vid=#{@vid}&#{$1}=#{$2}&#{base}#{$3}">#{$4}</a>)
			}
		end
	
		# դ href 򸡾ڤĤġ˥塼Υ󥯤 # FIXME!!: comment
		def menulink(date, present, href, content, list = true)
			if date == present
				q = %Q(#{content})
			else
				q = savelink(%Q(<a href="#{href}">#{content}</a>))
			end
			q = list(q) if list
			q
		end
	end

	# ᥤ(ǿ)ιܥ()˥塼
	class BottomMenu
		include LastDaySearch

		attr_reader :format_type

		def initialize(vid, day, type, prt, uri, last_id)
			@vid, @day, @type, @emu, @uri, @last_id = vid, day, type, prt, uri, last_id
			if last_day_search(@vid) == @day
				@format_type = 'update'
			elsif @emu.emulating? && @emu.state.last == @day
				@format_type = 'emulate'
			else
				@format_type = 'next'
			end
		end

		def generate
			$logger.debug("BottomMenu.generate: #{[@vid, @day, @type].join(', ')}")

			case @format_type
			when 'update'
				ERB.new(Template['part/bottom_update']).result(binding)
			when 'emulate'
				ERB.new(Template['part/bottom_prt']).result(binding)
			else
				@next_day = @day.to_i + 1
				unless File.exist?("#{Software::LogDir + @vid}/day#{@next_day}.dat")
					@next_day = 'party'
				end
				ERB.new(Template['part/bottom_next']).result(binding)
			end
		end
	end

	# å body Фĥȥγǧ
	class MessageLinker
		class << self
		include LastDaySearch

		# match or nomatch ˻ѤޤǤΥɤ߹
		# today  /\d+/ or 'party'(ninjin)
		def set(vid, today, type)
			$logger.debug("MessageLinker.set: #{vid}, #{today}, #{type}")
			@@today, @@type = today, type

			@@last_day = last_day_search(vid)
			@@ids = Marshal.load(File.read(Software::LogDir + vid + '/log.keys'))
			if @@today == @@last_day
				$logger.debug("MessageLinker.set: LastDayLogIDs.new @@last_day => #{@@last_day}")
				@@ids.update(LastDayLogIDs.new(vid, link_type_num()))
			end
			type_reject!()
		end

		def type_reject!
			case @@type
			when 'white'
				@@ids.reject! {|key, value|
					%w(think whisper groan).each {|type| value.delete(type)}
					value == []
				}
			when 'whisper'
				@@ids.reject! {|key, value|
					%w(think groan).each {|type| value.delete(type)}
					value == []
				}
			end
		end

		# å body ơ a Ǥɲä body ֤
		# FIXME!! base_id
		def link(today, body, base_id)
			body.gsub(_anchor_regex) {
	            match = $1.dup
				anchors_data = _parse_anchor(today, match)
	
				if match =~ /^\s*(..|.[^:].)$/
					match
				elsif anchors_data == false
					match
	            else
					_link(anchors_data, today, base_id, match)
				end
			}
		end

		# a Ǥʸ(Match)
		def _matched(href, baseid, time, content)
			js =  %Q(onmouseover="popup(event, '#{time}');" )
			%Q(<a #{js}>#{content}</a>)
		end
	
		# a Ǥʸ(NoMatch)
		def _nomatch(target, content)
			%Q{<a class="nomatch" title="No Match(#{target})">#{content}</a>}
		end

		end # class << self


		# ǸդΥ ID  type  Hash
		class LastDayLogIDs < Hash
			def initialize(vid, link_type_num)
				dats = Dir.glob(Software::LogDir + "#{vid}/*.dat").sort
				File.read(dats.last).each {|line|
					#msg = Message.new(line.chomp.split('<>'))
					#self[msg.id] = msg.type
					msg = line.chomp.split('<>')
					(self[msg[link_type_num]] ||= []).push msg[2]
				}
			end
		end
	end

	class EuroMessageLinker < MessageLinker
		class << self

		def link_type_num
			0  # 'id'
		end

		# 䥿פΥ󥫡ɽ
		def _anchor_regex
			%r{(\&gt;\&gt;((?:\d+:)*)([\+\*\-]*)(\d+))}
		end

		# anchor_regex ɽ match ʸ day, id  Hash ֤
		def _parse_anchor(today, time_str)
			time_str =~ _anchor_regex
			types = {'' => 'say', '-' => 'think', '*' => 'whisper', '+' => 'groan'}
			type, d, id = types[$3.dup], $2.delete(':'), $4.dup

			d = today if d == ''
			{'day' => d, 'id' => type + id }
		end

		def _link(date, today, base_id, match)
			day, aid = date['day'], date['id']

			if @@ids["#{day}:#{aid}"] && day == today
				_matched("##{aid}", base_id, "#{day}:#{aid}", match)
			elsif @@ids["#{day}:#{aid}"]
				_matched("#{day}##{aid}", base_id, "#{day}:#{aid}", match)
			elsif @@ids[aid] && day == @@last_day
				_matched("##{aid}", base_id, aid, match)
			else
				_nomatch("#{day}:#{aid}", match)
			end
		end

		end  # class << self
	end

	class NinjinMessageLinker < MessageLinker
		class << self

		def link_type_num
			3  # 'time'
		end

		# ե󥫡ɽ
		def _anchor_regex
			rD = /[0-9󻰻͸ϻȬ]/
			rT = /[0-9󻰻͸ϻȬ]/
			rS = /[:;\.\, \-\/\[\(]/
			rDS = /[:; h]/
			rDAY = /[d|day|||D|DAY|ġãģ||]/
			rAMPM = /AM||PM|У||/i
			rSPACE = /[:; \-]/

			%r{(
				(#{rD}{0,1}?)(#{rDAY}{0,3})#{rS}{0,1}
				#{rSPACE}*
				(#{rAMPM}*)
				#{rSPACE}*
				(#{rT}{1,2})(#{rDS}{0,2})(#{rT}{1,2})(?:ʬ|m|)*
			)}ixo
		end

		# anchor_regex ɽ match ʸ day, hour, minute  Hash ֤
		def _parse_anchor(today, time_str)
			time_str =~ _anchor_regex
	
			d, h, m, date_flag, ampm, ds = $2.dup, $5.dup, $7.dup, $3.dup, '', $6.dup
			ampm = $4 if $4
	
			if date_flag == ''
				h = d + h
				d = ''
			end
			d = (today.to_i - 1).to_s if d == ''
	
			{'' => '0', '' => '1', '' => '2', '' => '3',
			 '' => '4', '' => '5', '' => '6', '' => '7',
			 '' => '8', '' => '9', '' => '0', '' => '1',
			 '' => '2', '' => '3', '' => '4', '' => '5',
			 'ϻ' => '6', '' => '7', 'Ȭ' => '8', '' => '9',
			 '' => '1'}.each {|zen, han|
				[d, h, m].each {|x|
					x.gsub!(/#{zen}/, han)
				}
			}

			if h.size == 1 and m.size == 1 and ds == ''
				return(false)
			end
		
			h = h.to_i + 12 if ampm =~ /(PM|У|)/
			h = h.to_i - 24 if h.to_i >= 24
			return(false) if h.to_i > 23
			h = h.to_s
			h = '0' + h if h.size == 1
			m = '0' + m if m.size == 1
		
			d = today if d == ''
			{'day' => d, 'hour' => h.to_s, 'minute' => m }
		end

		# å body  a Ǥɲä body ֤
		def _link(date, today, base_id, match)
			day, time = date['day'], "#{date['hour']}:#{date['minute']}"

			child_past = @@ids["#{day}:#{time}"]
			child_today = @@ids[time]

			if child_past && day == today
				id = child_past.first['id']
				_matched("#m#{id}", base_id, "#{day}:#{time}", match)
			elsif child_past
				id = child_past.first['id']
				_matched("#{day}#m#{id}", base_id, "#{day}:#{time}", match)
			elsif child_today && day == @@last_day
				id = child_today.first['id']
				_matched("#m#{id}", base_id, "#{day}:#{time}", match)
			else
				_nomatch("#{day}d#{time}", match)
			end
		end

		end  # class << self
	end

	# ꥢ륿(P/RT) (20061014 Almost)
	class RealTimeEmulate
		attr_reader :state  # ɤ߹ߺѤդ Array

		def initialize(vid)
			@vid = vid
			@fn = Software::LogDir + @vid + '/.emu'
			@emulating = File.exist?(@fn)
			@state = File.read(@fn).chomp.split(',') if @emulating
		end

		def emulating?
			@emulating
		end

		def start
			File.open(@fn, 'w') {|fh| fh.write('1') }
		end

		# 1ʤƿʤ᤿դ֤
		def forward
			ary = File.read(@fn).chomp.split(',')
			x = ary.last.succ
			if File.exist?(Software::LogDir + @vid + "/day#{x}.dat")
				ary << x
				File.open(@fn, 'w') {|fh| fh.write ary.join(',') }
			else  # party
				quit
				x = 'party'
			end
			x
		end

		def quit
			File.unlink(@fn)
		end
	end

	# ºݤ BBS  HEAD/GET/POST Ф
	class Action
		def initialize(vid, user, pass)
			case vid
			when /^O/ then @map = EuroCountryMap.new(vid)
			else
				@map = NinjinCountryMap.new(vid)
			end
			h =  %Q({"user_id"%3D>"#{user}"%2C%20)
			h << %Q("password"%3D>"#{pass}"})
			@head = {'User-Agent' => "#{Software::Code} (#{Software::Version})"}
			@head['Cookie'] = 'login=' + CGI.escape(h) if user != ''
		end

		def head
			r = nil
			$logger.debug("Action#head: #{@map['host']}, #{@map['path4head']}")
			Net::HTTP.start(@map['host']) {|http|
				r = http.head(@map['path4head'], @head)
			}
			r
		end

		def get(fp, path = false)
			p = @map['uri4get']
			p = "#{@map['uri4get']}&amp;#{path}" if path
			$logger.debug("Action#get: " + p)
			if @head.has_key?('Cookie')
				f = OpenURI.open_uri(p,
					'User-Agent' => @head['User-Agent'],
					'Cookie' => @head['Cookie']
				)
			else
				f = OpenURI.open_uri(p, 'User-Agent' => @head['User-Agent'] )
			end
			File.open(fp, 'w') {|fh| fh.write f.read }
			true
		end
	
		def post(message, style)
			res = nil
			msg = CGI.escape(NKF.nkf('-sx', message))
			data =  "vid=#{@map['vnum']}&cmd=write&pid=&skill="
			data << "&style=#{style}&message=#{msg}"
			Net::HTTP.start(@map['host']) {|http|
				res = http.post(@map['path4post'], data, @head)
			}
			# PASS/UID ǧΤߡȯ¤ȤϸƤʤ
			if res.code == '302'
				true
			else
			# FIXME!!: raise error 
				false
			end
		end
	end

	# ȯ Hash
	class SayTypeCounter < Hash
		# key б Hash()  self ɲ
		def add(key)
			self[key] = \
				{'say' => 0, 'think' => 0, 'whisper' => 0, \
				 'groan' => 0, 'sys' => 0, 'act' => 0}
		end

		# ƤΥפȯι
		def total(key)
			t = 0
			self[key].each {|k, v| t += v }
			t
		end
	end

	#  (20061014 Almost But needed ViewPort)
	class Config
		def initialize
			@cfg = {}
	
			if File.exist?(Software::ConfigFile)
				File.read(Software::ConfigFile).each {|line|
					ary = line.chomp.split(':')
					@cfg[ary[0].strip] = ary[1].strip
				}
			else
			end
		end
	
		def [](key)
			if @cfg.has_key?(key)
				@cfg[key]
			else
				default(key)
			end
		end
	
		def default(key)
			defaults = {\
				'cast_alive_color' => '',
				'cast_alive_bgcolor' => '',
				'cast_newtype_color' => '',
				'cast_newtype_bgcolor' => 'gray',
				'cast_attack_color' => '',
				'cast_attack_bgcolor' => 'salmon',
				'cast_exe_color' => 'white',
				'cast_exe_bgcolor' => 'black',
	
				'system_css' => 'std.css',
				'groan_img' => false,
				'font_size' => '',
				'compact_mode' => false,
			}
			if defaults.has_key?(key)
				defaults[key]
			else
				''
			end
		end
	
		def dump(cfg_hash)
			f = ''
			cfg_hash.each {|k, v| f << k + ': ' + v[0] + "\n" if k != 'mode' }
			File.open(Software::ConfigFile, 'w') {|fh| fh.write f }
		end
	end

	# ȯ
	class VoiceStatistics
		require 'jcode'

		attr_reader :msgs_cnt
		attr_reader :avg_cnt
		attr_reader :first_day

		def initialize(vid)
			@vid = vid
			@msgs_cnt = {}
			n = '1'
			@first_day = n.dup
            @pkeys = Library::Ninjin::Persons.keys.delete_if {|x| x =~ /01|99/ }
			@pkeys.each {|key| @msgs_cnt[key] = [[], [], []] }

			while File.exist?("#{Software::LogDir + @vid}/day#{n}.dat")
				say_cnt, density, total = anlyse_voices(n)
				say_cnt.keys.each {|key|
					@msgs_cnt[key][0] << say_cnt[key]
					@msgs_cnt[key][1] << density[key]
					@msgs_cnt[key][2] << total[key]
				}
				n.succ!
			end

			[0, 1, 2].each {|n|
				@msgs_cnt.keys.each {|key|
					@msgs_cnt[key][n] << arys_avg(@msgs_cnt[key][n])
				}
			}

			delete_nitems!(@msgs_cnt)
			@avg_cnt = generate_village_avg(@msgs_cnt)
		end

		def delete_nitems!(msgs_cnt)
			@pkeys.each {|key|
				msgs_cnt.delete(key) if msgs_cnt[key][0][0] == 0
			}
		end

		def generate_village_avg(msgs_cnt)
			avg = [[], [], []]
			msgs_cnt.each {|key, value|
				[0, 1, 2].each {|i|
					value[i].each_with_index {|n, index|
						if avg[i][index]
							avg[i][index] << n
						else
							avg[i][index] = [n]
						end
					}
				}
			}
			[0, 1, 2].each {|i|
				avg[i].each_with_index {|ary, index|
					avg[i][index] = arys_avg(ary)
				}
			}
			avg
		end

		# Ϳ줿 Integer Array  0 ȴʿѤ֤
		def arys_avg(ary)
			x = ary - [0]
			if x.size != 0
				i = 0; ary.each {|j| i += j }  # Ǥι
				j = ary.size - (ary.size - x.size)  # 0 ǿ
				i / j
			else
				0
			end
		end

		# day_id ܤΥγơȯ̩١ʸǼ Hash Array ֤
		def anlyse_voices(day_id)
			fp = "#{Software::LogDir + @vid}/day#{day_id}.dat"
            day_str = day_id + 'D'
            say_cnt, density, total = {}, {}, {}  # ̩١ʸ
            @pkeys.each {|key| [say_cnt, density, total].each {|hash| hash[key] = 0}}

            File.read(fp).each {|line|
                ary = line.chomp.split('<>')
                if ary[2] == 'say' && ary[1] != '01'  # gerd
                    say_cnt[ary[1]] += 1
                    total[ary[1]] += ary[4].gsub(/<br \/>/, '..').jlength
                end
            }

            total.each {|k, v| density[k]= v / say_cnt[k] if say_cnt[k] != 0 }

            [say_cnt, density, total]
		end
    end

	class VoiceSplitter
		require 'jcode'

		def initialize(vid = '', day = '', say_type = '', linker = nil)
			@vid, @day, @say_type, @linker = vid, day, say_type, linker
		end

		def self.split(body)
			new().split(body)
		end

		def split(body)
			str = ''
			body.each {|line| str << line.chomp + "\r\n" }
			ret = []
			until str == ''
				tmp, tmp_base = '', ''
				if str =~ /(.{0,200})/em
					tmp_base = $1.dup
					tmp = $1.strip
					if tmp =~ /((.*?\r\n){5})/m
						tmp = $1.dup
						tmp_base = tmp
					end
					if tmp =~ /(.*?\r\n\-\-\r\n)/m
						tmp = $1.dup
						tmp_base = tmp
					end
					ret << tmp
					str[0, tmp_base.length] = ''
				end
			end
			normalize!(ret)
			[convert2msgs(ret), generate_status(ret)]
		end

		def normalize!(ret)
			ret.collect! {|x|
				s = x.strip.gsub(/^\-\-\Z/, '').gsub(/^\-\-\r\n/, '')
				s.strip.gsub(/^\-\Z/, '').gsub(/^\-\r\n/, '').strip
			}
			ret.reject! {|x| x.strip == '' }
			ret
		end

		def convert2msgs(msgs_array)
			msgs = []
			msgs_array.each {|msg|
				m = Message.new
				m['content'] = msg.gsub(/\r\n/, '<br>')
				m['density'] = msg.jlength
				l = 0
				msg.gsub(/^./) { l += 1 }
				m['lines'] = l
				msgs << m
			}
			link!(msgs) if @linker
			msgs
		end

		def link!(msgs)
			case @say_type
			when 'say'
				type = 'white'
			when 'think'
				type = 'think'
			when 'whisper'
				type = 'whisper'
			end
			@linker.set(@vid, @day, @type)
			msgs.map! {|msg|
				msg['content'] = @linker.link(@day, msg['content'], '')
				msg
			}
			msgs
		end

		def generate_status(msg_array)
			cnt_ary = []
			msg_array.each {|x| cnt_ary << x.jlength }

			avg_base = 0
			n = msg_array.size
			cnt_ary.each {|c| avg_base += c }
			avg = avg_base / n
			status = " #{n.to_s} ȯ (̩: #{avg.to_s} / ʸ: #{avg_base.to_s}): "
			status << cnt_ary.join('/')
			status
		end
	end

	class WolfBBSCastGenerator
		attr_reader :out
	
		def initialize(vid)
			@vid = vid
			@phash = _parse_announces()
			@config = Config.new
	
			all = []
			get_base().each {|ar|
				ar[2] = person_state(ar[0])
				all << ar
			}

			all.reject! {|x| x[0] == '' }
			all.reject! {|x| x == nil }
			all.sort! {|a, b|
				if a[0] == ''
					-1
				elsif b[0] == ''
					1
				elsif a[2][0] != b[2][0]
					a[2][0] <=> b[2][0]
				else
					if a[2][1] != b[2][1]
						case a[2][1]
						when '' then 1
						when '' then -1
						else  # 跺
							case b[2][1]
							when '' then -1
							when '' then 1
							end
						end
					else
						a[3] <=> b[3]
					end
				end
			}
	
			ret = '|LEFT:~' + %w(  ü   ).join('|LEFT:~')+ "|h\n"
			all.each {|ary|
				name = ary.shift
				user_id = ary.shift
				status = ary.shift
				user_skill = ary.shift
				name_ary = []
				Library::Rolls.each {|k, v| name_ary = [v[0], name] if k == name }
				name = 'CENTER:' + name_ary.join('&br;')
			
				x = []
				x << roll(name_ary[0], name_ary[1])
				x << name
				x << '[[' + user_id + ']]'
				x << death(status[0], status[1])
				x << skill(user_skill)
				x << ''
				ret << '|' + x.join('|') + "|\n"
			}
	
			@out = ret
		end
	
		def skill(sk)
			case sk
			when '¼'
				"&ref(#{_skill_url('00')}, nolink, #{sk});"
			else
				skills = {\
					'ϵ' => ['01', 'ffcccc'],
					'ꤤ' => ['02', 'ccffff'],
					'ǽ' => ['03', 'ccccff'],
					'' => ['04', 'ffccff'],
					'' => ['05', 'ccffcc'],
					'ͭ' => ['06', 'ffffcc'],
					'ϥॹʹ' => ['07', 'cc9966'],
				}
				no = skills[sk][0]
				c = skills[sk][1]
		
				"#{_skill_color(c)}&ref(#{_skill_url(no)},nolink,#{sk});"
			end
		end
		
		def _skill_color(color)
			"BGCOLOR(##{color}):"
		end
		
		def _skill_url(no)
			if no == '07'
				"http://ninjin001.x0.com/wolfe/plugin_wolf/img/skill07.jpg"
			else
				"http://ninjinix.x0.com/wolf/plugin_wolf/img/skill#{no}.jpg"
			end
		end
		
		def roll(r, name)
			no = Library::Ninjin::Images[name]
			if no == 'face17.jpg'
				x = "&ref(http://ninjin001.x0.com/wolfe/plugin_wolf/img/#{no}"
			else
				x = "&ref(http://ninjinix.x0.com/wolf/plugin_wolf/img/#{no}"
			end
			x << ",#{r} #{name});"
		end
		
		def death(d, type)
			hash = {'' => 'attack', '跺' => 'exe', '' => 'newtype', '¸' => 'alive'}
			forground = @config["cast_#{hash[type]}_color"]
			background = @config["cast_#{hash[type]}_bgcolor"]
			state = type
			state = "#{d}&br;" + state if type != '¸'

			if forground == ''
				if background == ''
					"CENTER:#{state}"
				else
					"CENTER:BGCOLOR(#{background}):#{state}"
				end
			else
				if background == ''
					"CENTER:&color(#{forground}){#{state}};"
				else
					"CENTER:BGCOLOR(#{background}):&color(#{forground}){#{state}};"
				end
			end
		end
	
		def get_base
			roll = IO.readlines(Software::LogDir + @vid + '/' + 'roll')
			roll.pop
			roll.shift
			
			ary = []
			roll.each {|line|
				l = line.chomp.gsub(/<\/td>/, '').split('<td>').map {|x|
					x.gsub(/<.*?>/, '')
				}
				l.shift
				ary << l
			}
			ary
		end
	
		def _parse_announces()
			phash = {}
			Library::Ninjin::Persons.each {|key, value| phash[value]= [] }
	
			last = 0
			Dir.entries(Software::LogDir + @vid + '/').sort.each {|fp|
				if fp =~ /(party|day(\d+))\.ann$/ 
					day = $2.dup.to_i if $2
					s = File.read(Software::LogDir + @vid + '/' + fp)
					{'' => '跺', '' => '', '' => ''}.each {|dtype, dvalue|
						if s.index(dtype)
							s.gsub(/^#{dtype}(.*)/) {
								$1.dup.split(' ').each {|p|
									if fp =~ /party/
										phash[p]= [last, dvalue]
									else
										phash[p]= [day, dvalue]
									end
								}
							}
						end
					}
					last = day.to_i + 1
				end
			}
			['', 'ʤ'].each {|c| phash.delete(c) }
			phash['']= [2, '']
			phash.each {|k, v| phash[k] = [99, '¸'] if v == [] }
			phash
		end
		private :_parse_announces
	
		def person_state(person)
			@phash[person]
		end
	end

	############################################################ VIEW

	class Panic
		def initialize
			@res = CGI.new
		end

		def self.print
			new().response
		end

		def content
			x = "An error occured.\n\n"
			x << backtrace()
			x << logfile() if File.exist?(Software::LogFile)
			x << summary()
			x
		end

		def backtrace
			x = "Backtrace\n------------------------\n"
			x << "#{$!}\n\n"
			x << %Q(#{$!.backtrace.join("\n")}\n)
			x
		end

		def logfile
			x = "\n\n"
			x << "#{Software::Code} log(last 5 lines)\n------------------------\n"
			log = IO.readlines(Software::LogFile)
			for i in 1..5
				x << log[log.size-i]
			end
			x
		end

		def summary
			x = "\n\nSystem summary\n------------------------\n"
			x << "Ruby #{RUBY_VERSION} (#{RUBY_PLATFORM})\n"
			x << "#{@res.user_agent}\n"
			x << "#{Software::Name} version #{Software::Version} "
			x << "(with #{@res.server_software})"
			x
		end

		def response
			@res.out('text/plain') { content() }
		end
	end

	class ToolChain
		class << self

		def set(vid)
			if vid =~ /^O/
				@@linker = EuroMessageLinker
				@@images = Library::Euro::Images
				@@type = 'euro'
				@@link_type = 'id'
				@@prologue = '1'
				@@map = EuroCountryMap.new(vid)
			else
				@@linker = NinjinMessageLinker
				@@images = Library::Ninjin::Images
				@@type = 'ninjin'
				@@link_type = 'time'
				@@prologue = '0'
				@@map = NinjinCountryMap.new(vid)
			end
		end

		def type; @@type; end
		def linker; @@linker; end
		def imagemap; @@images; end
		def link_type; @@link_type; end
		def prologue; @@prologue; end
		def map; @@map; end

		end  # class << self
	end

	class Loader
		def initialize
			$logger.debug(": Welcome, Master? (version.#{Software::Version})")
			@req = CGI.new
			Dir.mkdir(Software::LogDir) unless File.exist?(Software::LogDir)
		end

		def self.boot
			begin
				Library.load('trinity.sys')
				new().boot
			rescue
				Panic.print
			end
		end

		def boot
			if @req.has_key?('mode')
				page = __send__ "exec_#{@req.get('mode')}"
			elsif @req.has_key?('vid') && @req.get('vid') != ''
				if File.exist?(Software::LogDir + @req.get('vid'))
					page = exec_standard
				else
					page = exec_download
				end
			else
				page = exec_index
			end

			page.response
		end

		def get_std_keys
			ary = ['vid', 'day', 'type', 'sword']
			hash = {'vid' => '', 'day' => ToolChain.prologue, 'type' => 'white', 'sword' => nil }
			ret = @req.get_values_with_default(ary, hash)
			ret[2] = 'all' if ret[1] == 'party' #FIXME!! euro
			ret
		end

		def exec_standard
			ToolChain.set(@req.get('vid'))
			vid, day, type, sword = get_std_keys

			prt = @req.get('prt')
			RealTimeEmulate.new(vid).start if prt

			compat_log_keys(vid)
			Page::Standard.new(vid, day, type, sword)
		end

		def exec_prt_plus
			ToolChain.set(@req.get('vid'))
			vid, day, type, sword = get_std_keys

			prt = @req.get('prt')
			d = RealTimeEmulate.new(vid).forward

			Page::Standard.new(vid, d, type, sword)
		end

		def exec_prt_end
			ToolChain.set(@req.get('vid'))
			vid, day, type, sword = get_std_keys

			prt = @req.get('prt')
			d = RealTimeEmulate.new(vid).forward
			RealTimeEmulate.new(vid).quit

			Page::Standard.new(vid, d, type, sword)
		end

		def exec_index
			announce = []
			ii = InitImages.sync(Software::TemplateDir + 'imagelist.sys')
			announce << Software::Version + 'Ѥβեޤ' if ii
			Page::Index.new(announce)
		end

		def exec_config; Page::Config.new; end

		def exec_config_update
			cfg = Trinity::Config.new
			cfg.dump(@req.params)
			Page::Redirect.new('./?mode=config')
		end

		def exec_status
			Page::VoiceStatistics.new(@req.get('vid'))
		end

		def exec_castgen
			Page::CastGeneration.new(@req.get('vid'))
		end

		def exec_download
			ToolChain.set(@req.get('vid'))
			vid, day, type, sword = get_std_keys
			Trinity::Downloader.update(vid)
			Page::Standard.new(vid, day, type, sword)
		end

		def exec_update
			ToolChain.set(@req.get('vid'))
			vid, day, type, sword = get_std_keys
			dl_result = Trinity::Downloader.update(vid)
			Page::Update.new(vid, day, type, @req.get('refer'), dl_result)
		end

		def exec_split
			ToolChain.set(@req.get('vid'))
			vid, day, type, sword = get_std_keys
			Page::Split.new(vid, day)
		end

		def exec_split_res
			body = NKF.nkf('-ex', @req.get('body'))
			Page::SplitScrap.new(body)
		end

		def exec_preview
			ToolChain.set(@req.get('vid'))
			vid, day, type, sword = get_std_keys
			if @req.has_key?('html')
				say_type = 'say'
			elsif @req.has_key?('html_think')
				say_type = 'think'
			elsif @req.has_key?('html_whisper')
				say_type = 'whisper'
			elsif @req.has_key?('html_groan')
				say_type = 'groan'
			end
			Page::Preview.new(vid, day, type, say_type, @req.get('body'))
		end

		def exec_post
			ToolChain.set(@req.get('vid'))
			vid, day, type, sword = get_std_keys
			rtype, body = @req.get('rtype'), @req.get('body')
			msgs, junk = VoiceSplitter.split(CGI.unescape(body))
			user, pass = get_identity(vid)

			act = Action.new(vid, user, pass)
			w, first = true, true
			msgs.each {|msg|
				if first
					first = false
				else
					sleep 19
				end
				w = act.post(msg.content.gsub(/<br>/, "\n"), type)
				break unless w
			}
			url = "./?vid=#{vid}&day=#{day}&type=#{rtype}#bottom"
			#Trinity::Downloader.update(vid)
			if w
				Page::Redirect.new(url)
			else
				Page::PostError.new(vid, url)
			end
		end

		def get_identity(vid)
			cfg = Config.new
			user, pass = '', ''
			case vid
			when /^C/
				user = cfg['user_id_c']
				pass = cfg['password_c']
			when /^F/
				user = cfg['user_id_f']
				pass = cfg['password_f']
			when /^O/
				user = cfg['user_id_o']
				pass = cfg['password_o']
			end
			[user, pass]
		end

		def exec_message_res  # by AJAX.
			ToolChain.set(@req.get('vid'))
			vid, day, type, sword = get_std_keys
			Page::MessageScrap.new(vid, day, type, @req.get('id'))
		end

		def exec_redirect
			if @req.get('vid')
				ToolChain.set(@req.get('vid'))
				vid, day, type, sword = get_std_keys
				url = "./?vid=#{vid}&day=#{day}&type=#{type}##{@req.get('url')}"
			else
				url = @req.get('url')
			end
			Page::Redirect.new(url)
		end

		def exec_reparse
			dir = Software::LogDir + @req.get('vid')
			Dir.entries(dir + '/').sort.each {|fn|
				case fn
				when '.'
				when '..'
				else
					File.unlink(dir + '/' + fn)
				end
			}
			Dir.unlink(dir)

			Downloader.new(@req.get('vid'))
			Page::Redirect.new(@req.get('refer'))
		end

		# ZeX => Trinity
		def compat_log_keys(vid)
			unless File.exist?(Software::LogDir + vid + '/log.keys')
				Logs.new(vid, ToolChain.link_type)
			end
		end
	end

	module Page
		class PageBase
			def header
				{\
					'type' => 'text/html',
					'charset' => 'euc-jp'
				}
			end

			def convertable?
				true
			end

			def convert
				klass = self.class.to_s.sub(/.*::/, '').downcase
				erb = ERB.new(Template[klass])
				erb.result(binding)
			end

			def response
				if convertable?
					content = convert()
				else
					content = body()
				end
				res = CGI.new
				res.out(header()) { content }
			end

			def body
				@viewport.to_html  # @viewport => ViewPort Object
			end
		end

		class Standard < PageBase
			def initialize(vid, day, type, sword)
				$logger.debug("Page::Standard.new: #{vid}, #{day}, #{type}, #{sword}.")
				@vid, @day, @type, @sword = vid, day, type, sword
				@config = Trinity::Config.new
				@fontsize = @config['font_size'] + 'px' if @config['font_size'] != ''
				construct()
			end

			def construct
				Header.set(@vid, @day, @type)
				@announce = Header.announce
				@title = Header.title
				@url = ToolChain.map['uri']

				case @day
				when 'party' then fn = 'party.dat'
				else
					fn = "day#{@day}.dat"
				end

				stream = File.read("#{Software::LogDir}/#{@vid}/#{fn}")
				@viewport = ViewPort::Standard.new(@vid, @day, @type, @sword, stream)

				prt = RealTimeEmulate.new(@vid)
				types = @viewport.types.delete_if{|x| x == @type }.unshift(@type)
				@mainmenu = MainMenu.generate(@vid, @day, @sword, prt,
					@viewport.voice_counter, types, @viewport.type_counter)
				bm = BottomMenu.new(@vid, @day, @type, prt, @url, @viewport.last_id)
				@bottommenu = bm.generate()

				if File.exist?("#{Software::LogDir}/#{@vid}/.party") && bm.format_type == 'update'
					construct_ep()
				end
			end

			def construct_ep
				if ToolChain.type == 'ninjin'
					@castgen = ERB.new(Template['part/castgeneration']).result(binding)
				end
				@roll = File.read(Software::LogDir + "#{@vid}/roll")
			end
		end

		class MessageScrap < PageBase
			def initialize(vid, day, type, target)
				$logger.debug("Page::MessageScrap.new(#{target}).")
				@viewport = ViewPort::MessageScrap.new(vid, day, type, target)
			end

			def convertable?
				false
			end

			def header
				{\
				'type' => 'text/plain',
				'charset' => 'euc-jp',
				'cache-control' => 'no-cache',
				'pragma' => 'no-cache'
				}
			end
		end

		class Update < PageBase
			def initialize(vid, day, type, refer, dl_result)
				$logger.debug("Page::Update.new(#{refer}).")
				if dl_result
					@res = "./?mode=redirect&vid=#{vid}&day=#{day}&type=#{type}&url=#{refer}"
				else
					@res = 'no'
				end
			end

			def convertable?
				false
			end

			def header
				{\
				'type' => 'text/plain',
				'charset' => 'euc-jp',
				'cache-control' => 'no-cache',
				'pragma' => 'no-cache'
				}
			end

			def body
				@res
			end
		end

		class Index < PageBase
			def initialize(status)
				@status = status.join('<br>')
				$logger.debug("Page::Index.new.")
				@viewport = ViewPort::Index.new
			end
		end

		class Config < PageBase
			def initialize
				$logger.debug("Page::Config.new.")
				@viewport = ViewPort::Config.new
			end
		end

		class VoiceStatistics < PageBase
			def initialize(vid)
				@vid = vid
				$logger.debug("Page::VoiceStatistics.new(#{vid}).")
				@viewport = ViewPort::VoiceStatistics.new(vid)
			end
		end

		class Split < PageBase
			def initialize(vid, type)
				@vid, @type = vid, type
				$logger.debug("Page::Split.new.")
				@viewport = ViewPort::Split.new(vid, type)
			end
		end

		class SplitScrap < PageBase
			def initialize(body)
				$logger.debug("Page::SplitScrap.new.")
				@viewport = ViewPort::SplitScrap.new(body)
			end

			def convertable?
				false
			end
		end

		class Preview < PageBase
			def initialize(vid, day, type, say_type, body)
				@vid, @day, @type, @say_type = vid, day, type, say_type
				$logger.debug("Page::Preview.new.")
				@viewport = ViewPort::Preview.new(vid, day, type, say_type, body)
			end
		end

		class PostError < PageBase
			def initialize(vid, url)
				$logger.debug("Page::PostError.new.")
				@viewport = ViewPort::PostError.new(vid, url)
			end
		end

		class Redirect < PageBase
			def initialize(url)
				$logger.debug("Page::Redirect.new(#{url}).")
				@viewport = ViewPort::Redirect.new(url)
			end

			def convertable?
				false
			end
		end

		class CastGeneration < PageBase
			def initialize(vid)
				$logger.debug("Page::CastGeneration.new(#{vid}).")
				@generator = WolfBBSCastGenerator.new(vid)
			end

			def header
				{\
					'type' => 'text/plain',
					'charset' => 'euc-jp'
				}
			end

			def convertable?
				false
			end

			def body
				@generator.out
			end
		end
	end

	module ViewPort
		class ViewPortBase
			def to_html
				klass = self.class.to_s.sub(/.*::/, '').downcase
				erb = ERB.new(Template["viewport/#{klass}"])
				erb.result(binding)
			end
		end

		class MessageScrap < ViewPortBase
			include LastDaySearch

			def initialize(vid, day, type, target)
				@vid, @day, @type = vid, day, type
				@idtable = IDTable.new(vid)
				@idtable.load
				@config = Trinity::Config.new
				@linker = ToolChain.linker
				@linker.set(@vid, @day, @type)
				@plogs = Logs.new(vid, ToolChain.link_type, @type)
				@tlogs = []
				@last_day = last_day_search(@vid)
				if @day == @last_day
					day = 'day' + day unless @day == 'party'
					f = File.read(Software::LogDir + vid + "/#{day}.dat")
					f.each {|line| @tlogs << Message.new(line.chomp.split('<>'))}
				end
				@url = "./?vid=#{@vid}&day=#{target.split(':')[0]}&type=#{@type}"
				@msgs = get_msgs(target)
				@last_id = @msgs.last['id']
			end

			def get_msgs(target)
				msgs = []
				ary = target.split(':')
				target_day = ary.shift
				target_src = ary.join(':')
				# FIXME!! Euro&Ninjin compatible Anchor object.

				if target_day == @last_day
					targets = []
					@tlogs.each {|m| targets << m if m[ToolChain.link_type] == target_src }
				else
					targets = @plogs[target]
				end

				targets.each {|msg|
					face = ToolChain.imagemap[msg.person]
					if @config['groan_img'] == false && msg.icon == '99'
						face = 'face99.jpg'
					end
					face = 'null.png' unless face  # ˵Ѽ

					msgs << { \
						'id' => msg.id,
						'face' => face,
						'type' => msg.type,
						'body' => @linker.link(target_day, msg.content, msg.id),
						'sign' => msg.sign(target_day, '-')
						# FIXME!! Euro&Ninjin compatible Anchor object
					}
				}
				msgs
			end
		end

		class Index < ViewPortBase
			def initialize
				@config = Trinity::Config.new
				@vids = []
				Dir.entries(Software::LogDir).each {|f|
					if File.basename(f) =~ /\d+$/
						@vids << f if File.basename(f) != '0'
					end
				}
				@vids.sort!
			end 

			def user_id
				ary = []
				%w(c f o).each {|k|
					if @config["user_id_#{k}"] != ''
						ary << k.upcase + ': ' + @config["user_id_#{k}"]
					end
				}
				if ary != []
					'[ ' + ary.join(' / ') + ' ]'
				else
					''
				end
			end

			def village_state(vid)
				if File.exist?(Software::LogDir + vid + '/.emu')
					emu = RealTimeEmulate.new(vid)
					"P/RT: #{emu.state.last}"
				else
					path = Software::LogDir + vid + '/'
					dats = Dir.entries(path).delete_if{|dat| dat !~ /dat$/ }.sort

					begin
						i = IO.readlines(path + dats.first)
						if i[1] =~ /^\d{7,}/  # Ninjin
							i = i[1].split('<>').first
							Time.at(i.to_i).strftime("%Y/%m/%d")
						else
							i[1].split('<>')[3].split(' ')[0]
						end
					rescue
						''
					end
				end
			end
		end

		class Config < ViewPortBase
			def initialize
				@config = Trinity::Config.new
				@configset = Library::Config::All
				@cgi = CGI.new
			end
		end

		class Standard < ViewPortBase
			attr_reader :voice_counter, :types, :type_counter

			def initialize(vid, today, type, sword, stream)
				@msgs, @types = [], []
				@voice_counter = {}
				@type = type
				@config = Trinity::Config.new
				idtable = IDTable.new(vid)
				idtable.load
				saycnt = SayTypeCounter.new
				@type_counter = Hash.new {|hash, key| hash[key] = 0 }
				linker = ToolChain.linker
				linker.set(vid, today, type)

				stream.each {|line|
					msg =  Message.new(line.chomp.split('<>'))
					@types << msg.type unless @types.include?(msg.type)

					if by_type_include?(msg, today) && by_search_include?(msg, sword)
						saycnt.add(msg.person) unless saycnt.has_key?(msg.person)
						saycnt[msg.person][msg.type] += 1
	
						face = ToolChain.imagemap[msg.person]
						if @config['groan_img'] == false && msg.icon == '99'
							face = 'face99.jpg'
						end
						face = 'null.png' unless face  # ˵Ѽ

						@type_counter[msg.type] += 1
						tid = msg.type + @type_counter[msg.type].to_s
	
						fid = idtable.r[msg.person]
						if fid
							@voice_counter[fid] = 0 unless @voice_counter.has_key?(fid)
							@voice_counter[fid] += 1
						end
	
						@msgs << { \
							'fid' => "#{fid}_#{saycnt.total(msg.person)}",
							'tid' => tid,
							'id' => msg.id,
							'face' => face,
							'type' => msg.type,
							'body' => linker.link(today, msg.content, msg.id),
							'sign' => msg.sign(today, saycnt[msg.person][msg.type])
						}
					end
				}
				@types[@types.index('say')] = 'white' if @types.index('say')
			end

			def last_id
				if @msgs.last
					@msgs.last['id']
				else
					'top'
				end
			end

			def by_search_include?(msg, sword)
				return(true) unless sword
				return(true) if sword == ''

				regex = Regexp.compile(sword)
				if msg.content !~ regex
					false
				else
					msg.content.gsub!(regex) { '<em class="hilight">' + $&.dup + '</em>' }
				end
			end

			def by_type_include?(msg, today)
				if @type
					case @type
					when 'white' 
						if msg.type =~ /(whisper|think|groan)/o
							false
						elsif msg.type == 'sys' && today !~ /[01]/o
							false
						else
							true
						end
					when 'whisper'
						if msg.type =~ /(think|groan)/o
							false
						elsif msg.type == 'sys'
							if today !~ /[01]/o && msg.content !~ /¼ϵ/o
								false
							else
								true
							end
						else
							true
						end
					else
						true
					end
				end
			end
		end

		class VoiceStatistics < ViewPortBase
			def initialize(vid)
				@vid = vid
				status = Trinity::VoiceStatistics.new(@vid)
				@msgs_cnt = status.msgs_cnt
				@avg_cnt = status.avg_cnt
				@first_day = status.first_day
			end
		end

		class Split < ViewPortBase
			def initialize(vid, day)
				@vid, @day = vid, day
			end
		end

		class SplitScrap < ViewPortBase
			def initialize(body)
				@msgs, @status = VoiceSplitter.split(body)
			end
		end

		class Preview < ViewPortBase
			def initialize(vid, day, type, say_type, body)
				@vid, @day, @type, @say_type = vid, day, type, say_type
				vs = VoiceSplitter.new(vid, day, say_type, ToolChain.linker)
				@msgs, @status = vs.split(body)
				case @say_type
				when 'say' then @submit_label = 'ȯ'
				when 'think' then @submit_label = 'Ȥ'
				when 'whisper' then @submit_label = ''
				end
				@post_limit = @msgs.size * 20
				@bodyorig = CGI.escape(body)
			end
		end

		class PostError < ViewPortBase
			def initialize(vid, url)
				@vid, @url = vid, url
			end
		end

		class Redirect < ViewPortBase
			def initialize(url)
				@url = url
			end
		end

	end  # end of ViewPort Module
end

Trinity::Loader.boot if $0 == __FILE__
